import os
import shutil
import subprocess
from pathlib import Path
from typing import Optional

import requests
from truss.base.truss_config import ExternalData

B10CP_EXECUTABLE_NAME = "b10cp"
BLOB_DOWNLOAD_TIMEOUT_SECS = 600  # 10 minutes
B10CP_PATH_TRUSS_ENV_VAR_NAME = "B10CP_PATH_TRUSS"


def download_external_data(external_data: Optional[ExternalData], data_dir: Path):
    if external_data is None:
        return
    data_dir.mkdir(exist_ok=True)
    b10cp_path = _b10cp_path()

    # ensure parent directories exist
    for item in external_data.items:
        path = data_dir / item.local_data_path
        if data_dir not in path.parents:
            raise ValueError(
                "Local data path of external data cannot point to outside data directory"
            )
        path.parent.mkdir(exist_ok=True, parents=True)

    if b10cp_path is not None:
        print("b10cp found, using it to download external data")
        _download_external_data_using_b10cp(b10cp_path, data_dir, external_data)
        return

    # slow path
    _download_external_data_using_requests(data_dir, external_data)


def _b10cp_path() -> Optional[str]:
    return os.environ.get(B10CP_PATH_TRUSS_ENV_VAR_NAME)


def _download_external_data_using_b10cp(
    b10cp_path: str, data_dir: Path, external_data: ExternalData
):
    procs = []
    # TODO(pankaj) Limit concurrency here
    for item in external_data.items:
        path = (data_dir / item.local_data_path).resolve()
        proc = _download_from_url_using_b10cp(b10cp_path, item.url, path)
        procs.append(proc)

    for proc in procs:
        proc.wait()


def _download_from_url_using_b10cp(b10cp_path: str, url: str, download_to: Path):
    return subprocess.Popen(
        [
            b10cp_path,
            "-source",
            url,  # Add quotes to work with any special characters.
            "-target",
            str(download_to),
        ]
    )


def _download_external_data_using_requests(data_dir: Path, external_data: ExternalData):
    for item in external_data.items:
        download_from_url_using_requests(
            item.url, (data_dir / item.local_data_path).resolve()
        )


def download_from_url_using_requests(URL: str, download_to: Path):
    # Streaming download to keep memory usage low
    resp = requests.get(
        URL, allow_redirects=True, stream=True, timeout=BLOB_DOWNLOAD_TIMEOUT_SECS
    )
    resp.raise_for_status()
    with download_to.open("wb") as file:
        shutil.copyfileobj(resp.raw, file)



import pathlib
import sys
from enum import Enum
from typing import Optional

import pydantic
import truss


class DeployedChainlet(pydantic.BaseModel):
    name: str
    is_entrypoint: bool
    is_draft: bool
    status: str
    logs_url: str
    oracle_name: str


class ChainletArtifact(pydantic.BaseModel):
    truss_dir: pathlib.Path
    display_name: str
    name: str


class ModelOrigin(Enum):
    BASETEN = "BASETEN"
    CHAINS = "CHAINS"


class OracleData(pydantic.BaseModel):
    class Config:
        protected_namespaces = ()

    model_name: str
    s3_key: str
    encoded_config_str: str
    semver_bump: Optional[str] = "MINOR"
    version_name: Optional[str] = None


# This corresponds to `ChainletInputAtomicGraphene` in the backend.
class ChainletDataAtomic(pydantic.BaseModel):
    name: str
    oracle: OracleData


class TrussUserEnv(pydantic.BaseModel):
    truss_client_version: str
    python_version: str
    pydantic_version: str
    mypy_version: Optional[str]

    @classmethod
    def collect(cls):
        py_version = sys.version_info
        try:
            import mypy.version

            mypy_version = mypy.version.__version__
        except ImportError:
            mypy_version = None

        return cls(
            truss_client_version=truss.version(),
            python_version=f"{py_version.major}.{py_version.minor}.{py_version.micro}",
            pydantic_version=pydantic.version.version_short(),
            mypy_version=mypy_version,
        )



import asyncio
from typing import Any, Dict

import httpx
from fastapi import APIRouter
from fastapi.responses import JSONResponse, StreamingResponse
from helpers.errors import ModelLoadFailed, ModelNotReady
from httpx import URL, ConnectError, RemoteProtocolError
from starlette.requests import ClientDisconnect, Request
from starlette.responses import Response
from tenacity import RetryCallState, Retrying, retry_if_exception_type, wait_fixed

INFERENCE_SERVER_START_WAIT_SECS = 60

control_app = APIRouter()


@control_app.get("/")
def index():
    return {}


async def proxy(request: Request):
    inference_server_process_controller = (
        request.app.state.inference_server_process_controller
    )
    client: httpx.AsyncClient = request.app.state.proxy_client

    path = _reroute_if_health_check(request.url.path)
    url = URL(path=path, query=request.url.query.encode("utf-8"))

    # 2 min connect timeouts, no timeout for requests.
    # We don't want requests to fail due to timeout on the proxy
    timeout = httpx.Timeout(None, connect=2 * 60.0)
    try:
        request_body = await request.body()
    except ClientDisconnect:
        # If the client disconnects, we don't need to proxy the request
        return Response(status_code=499)

    inf_serv_req = client.build_request(
        request.method,
        url,
        headers=request.headers.raw,
        content=request_body,
        timeout=timeout,
    )

    # Wait a bit for inference server to start
    for attempt in Retrying(
        retry=(
            retry_if_exception_type(ConnectError)
            | retry_if_exception_type(ModelNotReady)
            | retry_if_exception_type(RemoteProtocolError)
            | retry_if_exception_type(httpx.ReadError)
            | retry_if_exception_type(httpx.ReadTimeout)
            | retry_if_exception_type(httpx.ConnectTimeout)
        ),
        stop=_custom_stop_strategy,
        wait=wait_fixed(1),
        reraise=False,
    ):
        with attempt:
            try:
                if inference_server_process_controller.is_inference_server_intentionally_stopped():
                    raise ModelLoadFailed("Model load failed")
                resp = await client.send(inf_serv_req, stream=True)

                if await _is_model_not_ready(resp):
                    raise ModelNotReady("Model has started running, but not ready yet.")
            except (RemoteProtocolError, ConnectError) as exp:
                # This check is a bit expensive so we don't do it before every request, we
                # do it only if request fails with connection error. If the inference server
                # process is running then we continue waiting for it to start (by retrying),
                # otherwise we bail.
                if (
                    inference_server_process_controller.inference_server_ever_started()
                    and not inference_server_process_controller.is_inference_server_running()
                ):
                    error_msg = (
                        "It appears your model has stopped running. This often means' \
                        ' it crashed and may need a fix to get it running again."
                    )
                    return JSONResponse(error_msg, 503)
                raise exp

    if _is_streaming_response(resp):
        return StreamingResponse(
            resp.aiter_bytes(), media_type="application/octet-stream"
        )

    await resp.aread()
    response = Response(resp.content, resp.status_code, resp.headers)
    return response


control_app.add_route("/v1/{path:path}", proxy, ["GET", "POST"])


@control_app.post("/control/patch")
async def patch(request: Request) -> Dict[str, str]:
    request.app.state.logger.info("Patch request received.")
    patch_request = await request.json()
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(
        None,
        lambda: request.app.state.inference_server_controller.apply_patch(
            patch_request
        ),
    )
    request.app.state.logger.info("Patch applied successfully")
    return {"msg": "Patch applied successfully"}


@control_app.get("/control/truss_hash")
async def truss_hash(request: Request) -> Dict[str, Any]:
    t_hash = request.app.state.inference_server_controller.truss_hash()
    return {"result": t_hash}


@control_app.post("/control/restart_inference_server")
async def restart_inference_server(request: Request) -> Dict[str, str]:
    request.app.state.inference_server_controller.restart()

    return {"msg": "Inference server started successfully"}


@control_app.get("/control/has_partially_applied_patch")
async def has_partially_applied_patch(request: Request) -> Dict[str, Any]:
    app_has_partially_applied_patch = (
        request.app.state.inference_server_controller.has_partially_applied_patch()
    )
    return {"result": app_has_partially_applied_patch}


@control_app.post("/control/stop_inference_server")
async def stop_inference_server(request: Request) -> Dict[str, str]:
    request.app.state.inference_server_controller.stop()
    return {"msg": "Inference server stopped successfully"}


async def _is_model_not_ready(resp) -> bool:
    if resp.status_code == 503:
        await resp.aread()
        return resp.content is not None and "model is not ready" in resp.content.decode(
            "utf-8"
        )
    return False


def _is_streaming_response(resp) -> bool:
    for header_name, value in resp.headers.items():
        if header_name.lower() == "transfer-encoding" and value.lower() == "chunked":
            return True
    return False


def _reroute_if_health_check(path: str) -> str:
    """
    Reroutes calls from the Operator to the inference server's health check endpoint (/v1/models/model) to /v1/models/model/loaded instead.
    This is done to avoid running custom health checks when the Operator is checking if the inference server is ready.
    """
    if path == "/v1/models/model":
        path = "/v1/models/model/loaded"
    return path


def _custom_stop_strategy(retry_state: RetryCallState) -> bool:
    # Stop after 10 attempts for ModelNotReady
    if retry_state.outcome is not None and isinstance(
        retry_state.outcome.exception(), ModelNotReady
    ):
        # Check if the retry limit for ModelNotReady has been reached
        return retry_state.attempt_number >= 10
    # For all other exceptions, stop after INFERENCE_SERVER_START_WAIT_SECS
    seconds_since_start = (
        retry_state.seconds_since_start
        if retry_state.seconds_since_start is not None
        else 0.0
    )
    return seconds_since_start >= INFERENCE_SERVER_START_WAIT_SECS



import logging
import subprocess
from pathlib import Path
from typing import Optional

from helpers.custom_types import (
    Action,
    ConfigPatch,
    EnvVarPatch,
    ExternalDataPatch,
    ModelCodePatch,
    PackagePatch,
    Patch,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from helpers.errors import UnsupportedPatch
from helpers.truss_patch.model_code_patch_applier import apply_code_patch

# TODO(marius/TaT): remove try-except after TaT.
try:
    from truss.base.truss_config import ExternalData, ExternalDataItem, TrussConfig
except ImportError:
    from truss.truss_config import (  # type: ignore[no-redef]
        ExternalData,
        ExternalDataItem,
        TrussConfig,
    )

from truss.util.download import download_external_data


class ModelContainerPatchApplier:
    """Applies patches to container running a truss.
    This should be compatible with TrussDirPatchApplier.
    """

    def __init__(
        self,
        inference_server_home: Path,
        app_logger: logging.Logger,
        pip_path: Optional[str] = None,  # Only meant for testing
    ) -> None:
        self._inference_server_home = inference_server_home
        self._model_module_dir = (
            self._inference_server_home / self._truss_config.model_module_dir
        )
        self._bundled_packages_dir = (
            self._inference_server_home / ".." / self._truss_config.bundled_packages_dir
        ).resolve()
        self._data_dir = self._inference_server_home / self._truss_config.data_dir
        self._app_logger = app_logger
        self._pip_path_cached = None
        if pip_path is not None:
            self._pip_path_cached = "pip"

    def __call__(self, patch: Patch, inf_env: dict):
        self._app_logger.debug(f"Applying patch {patch.to_dict()}")
        if isinstance(patch.body, ModelCodePatch):
            model_code_patch: ModelCodePatch = patch.body
            apply_code_patch(self._model_module_dir, model_code_patch, self._app_logger)
        elif isinstance(patch.body, PythonRequirementPatch):
            py_req_patch: PythonRequirementPatch = patch.body
            self._apply_python_requirement_patch(py_req_patch)
        elif isinstance(patch.body, SystemPackagePatch):
            sys_pkg_patch: SystemPackagePatch = patch.body
            self._apply_system_package_patch(sys_pkg_patch)
        elif isinstance(patch.body, ConfigPatch):
            config_patch: ConfigPatch = patch.body
            self._apply_config_patch(config_patch)
        elif isinstance(patch.body, EnvVarPatch):
            env_var_patch: EnvVarPatch = patch.body
            self._apply_env_var_patch(env_var_patch, inf_env)
        elif isinstance(patch.body, ExternalDataPatch):
            external_data_patch: ExternalDataPatch = patch.body
            self._apply_external_data_patch(external_data_patch)
        elif isinstance(patch.body, PackagePatch):
            package_patch: PackagePatch = patch.body
            apply_code_patch(
                self._bundled_packages_dir, package_patch, self._app_logger
            )
        else:
            raise UnsupportedPatch(f"Unknown patch type {patch.type}")

    @property
    def _truss_config(self) -> TrussConfig:
        return TrussConfig.from_yaml(self._inference_server_home / "config.yaml")

    @property
    def _pip_path(self) -> str:
        if self._pip_path_cached is None:
            self._pip_path_cached = _identify_pip_path()
        return self._pip_path_cached

    def _apply_python_requirement_patch(
        self, python_requirement_patch: PythonRequirementPatch
    ):
        self._app_logger.debug(
            f"Applying python requirement patch {python_requirement_patch.to_dict()}"
        )
        action = python_requirement_patch.action

        if action == Action.REMOVE:
            subprocess.run(
                [
                    self._pip_path,
                    "uninstall",
                    "-y",
                    python_requirement_patch.requirement,
                ],
                check=True,
            )
        elif action in [Action.ADD, Action.UPDATE]:
            subprocess.run(
                [
                    self._pip_path,
                    "install",
                    python_requirement_patch.requirement,
                    "--upgrade",
                ],
                check=True,
            )
        else:
            raise ValueError(f"Unknown python requirement patch action {action}")

    def _apply_system_package_patch(self, system_package_patch: SystemPackagePatch):
        self._app_logger.debug(
            f"Applying system package patch {system_package_patch.to_dict()}"
        )
        action = system_package_patch.action

        if action == Action.REMOVE:
            subprocess.run(
                ["apt", "remove", "-y", system_package_patch.package], check=True
            )
        elif action in [Action.ADD, Action.UPDATE]:
            subprocess.run(["apt", "update"], check=True)
            subprocess.run(
                ["apt", "install", "-y", system_package_patch.package], check=True
            )
        else:
            raise ValueError(f"Unknown python requirement patch action {action}")

    def _apply_config_patch(self, config_patch: ConfigPatch):
        self._app_logger.debug(f"Applying config patch {config_patch.to_dict()}")
        TrussConfig.from_dict(config_patch.config).write_to_yaml_file(
            Path(self._inference_server_home / config_patch.path)
        )

    def _apply_env_var_patch(self, env_var_patch: EnvVarPatch, inf_env: dict):
        self._app_logger.debug(
            f"Applying environment variable patch {env_var_patch.to_dict()}"
        )
        action = env_var_patch.action
        ((env_var_name, env_var_value),) = env_var_patch.item.items()

        if action == Action.REMOVE:
            inf_env.pop(env_var_name, None)
        elif action in [Action.ADD, Action.UPDATE]:
            inf_env.update({env_var_name: env_var_value})
        else:
            raise ValueError(f"Unknown patch action {action}")

    def _apply_external_data_patch(self, external_data_patch: ExternalDataPatch):
        self._app_logger.debug(
            f"Applying external data patch {external_data_patch.to_dict()}"
        )
        action = external_data_patch.action
        if action == Action.REMOVE:
            item = ExternalDataItem.from_dict(external_data_patch.item)
            filepath = self._data_dir / item.local_data_path
            if not filepath.exists():
                self._app_logger.warning(
                    f"Could not delete file {filepath}: not found."
                )
            else:
                self._app_logger.debug(f"Deleting file {filepath}")
                filepath.unlink()
        elif action == Action.ADD:
            download_external_data(
                ExternalData.from_list([external_data_patch.item]), self._data_dir
            )
        else:
            raise ValueError(f"Unknown patch action {action}")


def _identify_pip_path() -> str:
    if Path("/usr/local/bin/pip3").exists():
        return "/usr/local/bin/pip3"

    if Path("/usr/local/bin/pip").exists():
        return "/usr/local/bin/pip"

    raise RuntimeError("Unable to find pip, make sure it's installed.")



import base64
import contextlib
import json
import os
from typing import TYPE_CHECKING, Optional, Type

import boto3
from boto3.s3.transfer import TransferConfig
from truss.util.env_vars import override_env_vars

if TYPE_CHECKING:
    from rich import progress


def base64_encoded_json_str(obj):
    return base64.b64encode(str.encode(json.dumps(obj))).decode("utf-8")


def multipart_upload_boto3(
    file_path,
    bucket_name: str,
    key: str,
    credentials: dict,
    progress_bar: Optional[Type["progress.Progress"]],
) -> None:
    # In the CLI flow, ignore any local ~/.aws/config files,
    # which can interfere with uploading the Truss to S3.
    with override_env_vars({"AWS_CONFIG_FILE": ""}):
        s3_resource = boto3.resource("s3", **credentials)
        filesize = os.stat(file_path).st_size

        progress_context = (
            progress_bar(transient=True) if progress_bar else contextlib.nullcontext()
        )
        task_id = (
            progress_context.add_task("[cyan]Uploading Truss", total=filesize)
            if not isinstance(progress_context, contextlib.nullcontext)
            else None
        )

        def callback(bytes_transferred):
            if progress_bar:
                progress_context.update(task_id, advance=bytes_transferred)

        with progress_context:
            s3_resource.Object(bucket_name, key).upload_file(
                file_path,
                Config=TransferConfig(max_concurrency=10, use_threads=True),
                Callback=callback,
            )



from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, model_validator


class ModelInput(BaseModel):
    """This class mirrors the `CompletionCreateParamsBase` in the `openai-python` repository.

    However, that class is a TypedDict rather than a pydantic model, so we redefine it here
    to take advantage of pydantic's validation features. In addition, we define helper methods
    to get the formatted prompt, tools to use, and response format to adhere to.

    Unsupported parameters:
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-store
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-metadata
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-frequency_penalty
      - Frequency penalty is not currently passed through to briton
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-logit_bias
      - User provided logit biasing is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-logprobs
      - Returning log probabilities is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-top_logprobs
      - Returning log probabilities is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-service_tier
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-user
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-function_call
      - Deprecated
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-functions
      - Deprecated
    """

    class Tool(BaseModel):
        """An element in the top level `tools` field."""

        class Function(BaseModel):
            name: str
            description: Optional[str] = Field(None)
            parameters_: Optional[Dict[str, Any]] = Field(None, alias="parameters")
            return_: Optional[Dict[str, Any]] = Field(None, alias="return")

            @model_validator(mode="after")
            def definitions_valid(cls, values):
                if "definitions" in values.parameters and "$defs" in values.parameters:
                    raise ValueError(
                        "Both pydantic v1 and v2 definitions found; please check schema."
                    )
                return values

            @property
            def parameters(self) -> Dict[str, Any]:
                if self.parameters_ is None:
                    return {"properties": {}}
                elif "properties" not in self.parameters_:
                    return {"properties": {}, **self.parameters_}
                else:
                    return self.parameters_

            @property
            def parameters_without_definitions(self) -> Dict[str, Any]:
                parameters = self.parameters.copy()
                for keyword in ["definitions", "$defs"]:
                    parameters.pop(keyword, None)
                return parameters

            @property
            def definitions(self) -> Optional[tuple[Dict[str, Any], str]]:
                for keyword in ["definitions", "$defs"]:
                    if keyword in self.parameters:
                        return self.parameters[keyword], keyword
                return None

            @property
            def json_schema(self) -> Dict[str, Any]:
                return {
                    "type": "object",
                    "properties": {
                        "name": {"const": self.name},
                        "parameters": self.parameters_without_definitions,
                    },
                    "required": ["name", "parameters"],
                }

        type: Literal["function"]
        function: Function

    class ToolChoice(BaseModel):
        """The top level `tool_choice` field."""

        class FunctionChoice(BaseModel):
            name: str

        type: Literal["function"]
        function: FunctionChoice

    class SchemaResponseFormat(BaseModel):
        """The top level `response_format` field."""

        class JsonSchema(BaseModel):
            """`schema_` holds the actual json schema"""

            schema_: Dict[str, Any] = Field(..., alias="schema")

        type: Literal["json_schema"]
        json_schema: JsonSchema

    class JsonResponseFormat(BaseModel):
        type: Literal["json_object"]

    class TextResponseFormat(BaseModel):
        type: Literal["text"]

    class StreamOptions(BaseModel):
        """The top level `stream_options` field."""

        include_usage: bool

    class LookaheadDecodingConfig(BaseModel):
        window_size: int
        ngram_size: int
        verification_set_size: int

    model: Optional[str] = Field("")

    messages: Optional[List[Dict[str, Any]]] = Field(None)
    prompt: Optional[str] = Field(None, min_length=1)

    max_tokens: Optional[int] = Field(None)
    max_completion_tokens: Optional[int] = Field(None)

    stream: Optional[bool] = Field(None)
    stream_options: Optional[StreamOptions] = Field(None)

    seed: Optional[int] = Field(None)
    random_seed: Optional[int] = Field(None)

    frequency_penalty: Optional[float] = Field(0)
    presence_penalty: Optional[float] = Field(0)
    length_penalty: Optional[float] = Field(None)

    # Not part of openai spec but supported by briton
    repetition_penalty: Optional[float] = Field(None)
    temperature: Optional[float] = Field(1.0)
    top_p: Optional[float] = Field(1.0)
    runtime_top_p: Optional[float] = Field(None)
    top_k: Optional[int] = Field(50)
    runtime_top_k: Optional[int] = Field(None)
    stop: Optional[Union[str, List[str]]] = Field(None)
    bad_words_: Optional[Union[str, List[str]]] = Field(None)
    skip_special_tokens: Optional[List[str]] = Field(None)
    response_format: Optional[
        Union[SchemaResponseFormat, JsonResponseFormat, TextResponseFormat]
    ] = Field(None)
    tools: Optional[List[Tool]] = Field(None)
    tool_choice: Optional[Union[Literal["none", "required", "auto"], ToolChoice]] = (
        Field(None)
    )
    parallel_tool_calls: Optional[bool] = Field(True)
    beam_width: Optional[Literal[1]] = Field(None)
    n: Optional[int] = Field(1)
    end_id: Optional[int] = Field(None)
    pad_id: Optional[int] = Field(None)
    # WiM fields
    margins_prompt: Optional[str] = Field(None)
    margins_stop_sequences: Optional[List[str]] = Field(["NO#"])
    max_chunk_size: Optional[int] = Field(4096)
    # Lookahead Decoding
    lookahead_decoding_config: Optional[LookaheadDecodingConfig] = Field(None)



# TODO: this file contains too much implementation -> restructure.
import abc
import enum
import logging
import pathlib
import traceback
from typing import (  # type: ignore[attr-defined]  # Chains uses Python >=3.9.
    Any,
    AsyncIterator,
    Callable,
    ClassVar,
    Generic,
    GenericAlias,  # This causes above type error.
    Iterable,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
    runtime_checkable,
)

import pydantic
from truss.base import truss_config
from truss.base.constants import PRODUCTION_ENVIRONMENT_NAME

from truss_chains import baseten_llm

BASETEN_API_SECRET_NAME = "baseten_chain_api_key"
SECRET_DUMMY = "***"
TRUSS_CONFIG_CHAINS_KEY = "chains_metadata"
GENERATED_CODE_DIR = ".chains_generated"
DYNAMIC_CHAINLET_CONFIG_KEY = "dynamic_chainlet_config"
OTEL_TRACE_PARENT_HEADER_KEY = "traceparent"
RUN_REMOTE_METHOD_NAME = "run_remote"  # Chainlet method name exposed as endpoint.
MODEL_ENDPOINT_METHOD_NAME = "predict"  # Model method name exposed as endpoint.
HEALTH_CHECK_METHOD_NAME = "is_healthy"
# Below arg names must correspond to `definitions.ABCChainlet`.
CONTEXT_ARG_NAME = "context"  # Referring to Chainlets `__init__` signature.
SELF_ARG_NAME = "self"
REMOTE_CONFIG_NAME = "remote_config"

K = TypeVar("K", contravariant=True)
V = TypeVar("V", covariant=True)


C = TypeVar("C")


class _classproperty(Generic[C, V]):
    def __init__(self, fget: Callable[[Type[C]], V]) -> None:
        self._fget = fget

    def __get__(self, instance: object, owner: Type[C]) -> V:
        return self._fget.__get__(None, owner)()


def classproperty(fget: Callable[[Type[C]], V]) -> _classproperty[C, V]:
    return _classproperty(fget)


@runtime_checkable
class MappingNoIter(Protocol[K, V]):
    def __getitem__(self, key: K) -> V: ...

    def __len__(self) -> int: ...

    def __contains__(self, key: K) -> bool: ...


class SafeModel(pydantic.BaseModel):
    """Pydantic base model with reasonable config."""

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=False, strict=True, validate_assignment=True
    )


class SafeModelNonSerializable(pydantic.BaseModel):
    """Pydantic base model with reasonable config - allowing arbitrary types."""

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        strict=True,
        validate_assignment=True,
        extra="forbid",
    )


class ChainsUsageError(TypeError):
    """Raised when user-defined Chainlets do not adhere to API constraints."""


class MissingDependencyError(TypeError):
    """Raised when a needed resource could not be found or is not defined."""


class ChainsRuntimeError(Exception):
    """Raised when components are not used the expected way at runtime."""


class ChainsDeploymentError(Exception):
    """Raised when interaction with a Chain deployment are not possible."""


class AbsPath:
    _abs_file_path: str
    _creating_module: str
    _original_path: str

    def __init__(
        self, abs_file_path: str, creating_module: str, original_path: str
    ) -> None:
        self._abs_file_path = abs_file_path
        self._creating_module = creating_module
        self._original_path = original_path

    def _raise_if_not_exists(self, abs_path: str) -> None:
        path = pathlib.Path(abs_path)
        if not (path.is_file() or (path.is_dir() and any(path.iterdir()))):
            raise MissingDependencyError(
                f"With the file path `{self._original_path}` an absolute path relative "
                f"to the calling module `{self._creating_module}` was created, "
                f"resulting `{self._abs_file_path}` - but no file was found."
            )

    @property
    def abs_path(self) -> str:
        if self._abs_file_path != self._original_path:
            logging.debug(
                f"Using abs path `{self._abs_file_path}` for path specified as "
                f"`{self._original_path}` (in `{self._creating_module}`)."
            )
        abs_path = self._abs_file_path
        self._raise_if_not_exists(abs_path)
        return abs_path


class BasetenImage(enum.Enum):
    """Default images, curated by baseten, for different python versions. If a Chainlet
    uses GPUs, drivers will be included in the image."""

    # Enum values correspond to truss canonical python versions.
    PY39 = "py39"
    PY310 = "py310"
    PY311 = "py311"


class CustomImage(SafeModel):
    """Configures the usage of a custom image hosted on dockerhub."""

    image: str
    python_executable_path: Optional[str] = None
    docker_auth: Optional[truss_config.DockerAuthSettings] = None


class DockerImage(SafeModelNonSerializable):
    """Configures the docker image in which a remoted chainlet is deployed.

    Note:
        Any paths are relative to the source file where ``DockerImage`` is
        defined and must be created with the helper function ``make_abs_path_here``.
        This allows you for example organize chainlets in different (potentially nested)
        modules and keep their requirement files right next their python source files.

    Args:
        base_image: The base image used by the chainlet. Other dependencies and
          assets are included as additional layers on top of that image. You can choose
          a Baseten default image for a supported python version (e.g.
          ``BasetenImage.PY311``), this will also include GPU drivers if needed, or
          provide a custom image (e.g. ``CustomImage(image="python:3.11-slim")``)..
        pip_requirements_file: Path to a file containing pip requirements. The file
          content is naively concatenated with ``pip_requirements``.
        pip_requirements: A list of pip requirements to install.  The items are
          naively concatenated with the content of the ``pip_requirements_file``.
        apt_requirements: A list of apt requirements to install.
        data_dir: Data from this directory is copied into the docker image and
          accessible to the remote chainlet at runtime.
        external_package_dirs: A list of directories containing additional python
          packages outside the chain's workspace dir, e.g. a shared library. This code
          is copied into the docker image and importable at runtime.
    """

    base_image: Union[BasetenImage, CustomImage] = BasetenImage.PY311
    pip_requirements_file: Optional[AbsPath] = None
    pip_requirements: list[str] = []
    apt_requirements: list[str] = []
    data_dir: Optional[AbsPath] = None
    external_package_dirs: Optional[list[AbsPath]] = None

    @pydantic.root_validator(pre=True)
    def migrate_fields(cls, values):
        if "base_image" in values:
            base_image = values["base_image"]
            if isinstance(base_image, str):
                doc_link = "https://docs.baseten.co/chains-reference/sdk#class-truss-chains-dockerimage"
                raise ChainsUsageError(
                    "`DockerImage.base_image` as string is deprecated. Specify as "
                    f"`BasetenImage` or `CustomImage` (see docs: {doc_link})."
                )
        return values


class ComputeSpec(pydantic.BaseModel):
    """Parsed and validated compute.  See ``Compute`` for more information."""

    # TODO[rcano] add node count
    cpu_count: int = 1
    predict_concurrency: int = 1
    memory: str = "2Gi"
    accelerator: truss_config.AcceleratorSpec = truss_config.AcceleratorSpec()


CpuCountT = Literal["cpu_count"]
CPU_COUNT: CpuCountT = "cpu_count"


class Compute:
    """Specifies which compute resources a chainlet has in the *remote* deployment.

    Note:
        Not all combinations can be exactly satisfied by available hardware, in some
        cases more powerful machine types are chosen to make sure requirements are met
        or over-provisioned. Refer to the
        `baseten instance reference <https://docs.baseten.co/performance/instances>`_.
    """

    # Builder to create ComputeSpec.
    # This extra layer around `ComputeSpec` is needed to parse the accelerator options.

    _spec: ComputeSpec

    def __init__(
        self,
        cpu_count: int = 1,
        memory: str = "2Gi",
        gpu: Union[str, truss_config.Accelerator, None] = None,
        gpu_count: int = 1,
        predict_concurrency: Union[int, CpuCountT] = 1,
    ) -> None:
        """
        Args:
            cpu_count: Minimum number of CPUs to allocate.
            memory: Minimum memory to allocate, e.g. "2Gi" (2 gibibytes).
            gpu: GPU accelerator type, e.g. "A10G", "A100", refer to the
              `truss config <https://docs.baseten.co/reference/config#resources-accelerator>`_
              for more choices.
            gpu_count: Number of GPUs to allocate.
            predict_concurrency: Number of concurrent requests a single replica of a
              deployed chainlet handles.

        Concurrency concepts are explained in `this guide <https://docs.baseten.co/deploy/guides/concurrency#predict-concurrency>`_. # noqa: E501
        It is important to understand the difference between `predict_concurrency` and
        the concurrency target (used for autoscaling, i.e. adding or removing replicas).
        Furthermore, the ``predict_concurrency`` of a single instance is implemented in
        two ways:

        - Via python's ``asyncio``, if ``run_remote`` is an async def. This
          requires that ``run_remote`` yields to the event loop.

        - With a threadpool if it's a synchronous function. This requires
          that the threads don't have significant CPU load (due to the GIL).
        """
        accelerator = truss_config.AcceleratorSpec()
        if gpu:
            accelerator.accelerator = truss_config.Accelerator(gpu)
            accelerator.count = gpu_count
            accelerator = truss_config.AcceleratorSpec(
                accelerator=truss_config.Accelerator(gpu), count=gpu_count
            )
        if predict_concurrency == CPU_COUNT:
            predict_concurrency_int = cpu_count
        else:
            assert isinstance(predict_concurrency, int)
            predict_concurrency_int = predict_concurrency

        self._spec = ComputeSpec(
            cpu_count=cpu_count,
            memory=memory,
            accelerator=accelerator,
            predict_concurrency=predict_concurrency_int,
        )

    def get_spec(self) -> ComputeSpec:
        return self._spec.model_copy(deep=True)


class AssetSpec(SafeModel):
    """Parsed and validated assets. See ``Assets`` for more information."""

    secrets: dict[str, str] = pydantic.Field({})
    cached: list[truss_config.ModelRepo] = []
    external_data: list[truss_config.ExternalDataItem] = []


class Assets:
    """Specifies which assets a chainlet can access in the remote deployment.

    For example, model weight caching can be used like this::

        import truss_chains as chains
        from truss.base import truss_config

        mistral_cache = truss_config.ModelRepo(
            repo_id="mistralai/Mistral-7B-Instruct-v0.2",
            allow_patterns=["*.json", "*.safetensors", ".model"]
        )
        chains.Assets(cached=[mistral_cache], ...)

    See `truss caching guide <https://docs.baseten.co/deploy/guides/model-cache#enabling-caching-for-a-model>`_
    for more details on caching.
    """

    # Builder to create asset spec.
    # This extra layer around `AssetSpec` is needed to add secret_keys.
    _spec: AssetSpec

    def __init__(
        self,
        cached: Iterable[truss_config.ModelRepo] = (),
        secret_keys: Iterable[str] = (),
        external_data: Iterable[truss_config.ExternalDataItem] = (),
    ) -> None:
        """
        Args:
            cached: One or more ``truss_config.ModelRepo`` objects.
            secret_keys: Names of secrets stored on baseten, that the
              chainlet should have access to. You can manage secrets on baseten
              `here <https://app.baseten.co/settings/secrets>`_.
            external_data: Data to be downloaded from public URLs and made available
              in the deployment (via ``context.data_dir``). See
              `here <https://docs.baseten.co/reference/config#external-data>`_ for
              more details.
        """
        self._spec = AssetSpec(
            cached=list(cached),
            secrets={k: SECRET_DUMMY for k in secret_keys},
            external_data=list(external_data),
        )

    def get_spec(self) -> AssetSpec:
        """Returns parsed and validated assets."""
        return self._spec.model_copy(deep=True)


class ChainletOptions(SafeModelNonSerializable):
    """
    Args:
        enable_b10_tracing: enables baseten-internal trace data collection. This
          helps baseten engineers better analyze chain performance in case of issues.
          It is independent of a potentially user-configured tracing instrumentation.
          Turning this on, could add performance overhead.
        env_variables: static environment variables available to the deployed chainlet.
        health_checks: Configures health checks for the chainlet.
    """

    enable_b10_tracing: bool = False
    env_variables: Mapping[str, str] = {}
    health_checks: truss_config.HealthChecks = truss_config.HealthChecks()


class ChainletMetadata(SafeModelNonSerializable):
    is_entrypoint: bool = False
    chain_name: Optional[str] = None
    init_is_patched: bool = False


class EntityType(str, enum.Enum):
    CHAINLET = enum.auto()
    MODEL = enum.auto()
    BASETEN_LLM = enum.auto()


class FrameworkConfig(SafeModelNonSerializable):
    entity_type: EntityType
    supports_dependencies: bool
    endpoint_method_name: str


class RemoteConfig(SafeModelNonSerializable):
    """Bundles config values needed to deploy a chainlet remotely.

    This is specified as a class variable for each chainlet class, e.g.::

            import truss_chains as chains


            class MyChainlet(chains.ChainletBase):
                remote_config = chains.RemoteConfig(
                    docker_image=chains.DockerImage(
                        pip_requirements=["torch==2.0.1", ...]
                    ),
                    compute=chains.Compute(cpu_count=2, gpu="A10G", ...),
                    assets=chains.Assets(secret_keys=["hf_access_token"], ...),
                )

    """

    docker_image: DockerImage = DockerImage()
    compute: Compute = Compute()
    assets: Assets = Assets()
    name: Optional[str] = None
    options: ChainletOptions = ChainletOptions()

    def get_compute_spec(self) -> ComputeSpec:
        return self.compute.get_spec()

    def get_asset_spec(self) -> AssetSpec:
        return self.assets.get_spec()


DEFAULT_TIMEOUT_SEC = 600.0


class RPCOptions(SafeModel):
    """Options to customize RPCs to dependency chainlets.

    Args:
        retries: The number of times to retry the remote chainlet in case of failures
          (e.g. due to transient network issues). For streaming, retries are only made
          if the request fails before streaming any results back. Failures mid-stream
          not retried.
        timeout_sec: Timeout for the HTTP request to this chainlet.
        use_binary: Whether to send data in binary format. This can give a parsing
         speedup and message size reduction (~25%) for numpy arrays. Use
         ``NumpyArrayField`` as a field type on pydantic models for integration and set
         this option to ``True``. For simple text data, there is no significant benefit.
    """

    retries: int = 1
    timeout_sec: float = DEFAULT_TIMEOUT_SEC
    use_binary: bool = False


class ServiceDescriptor(SafeModel):
    """Bundles values to establish an RPC session to a dependency chainlet,
    specifically with ``StubBase``."""

    name: str
    display_name: str
    options: RPCOptions


class DeployedServiceDescriptor(ServiceDescriptor):
    predict_url: str


class Environment(SafeModel):
    """The environment the chainlet is deployed in.

    Args:
        name: The name of the environment.
    """

    name: str
    # can add more fields here as we add them to dynamic_config configmap


class DeploymentContext(SafeModelNonSerializable):
    """Bundles config values and resources needed to instantiate Chainlets.

    The context can optionally added as a trailing argument in a Chainlet's
    ``__init__`` method and then used to set up the chainlet (e.g. using a secret as
    an access token for downloading model weights).

    Args:
        data_dir: The directory where the chainlet can store and access data,
          e.g. for downloading model weights.
        chainlet_to_service: A mapping from chainlet names to service descriptors.
          This is used to create RPC sessions to dependency chainlets. It contains only
          the chainlet services that are dependencies of the current chainlet.
        secrets: A mapping from secret names to secret values. It contains only the
          secrets that are listed in ``remote_config.assets.secret_keys`` of the
          current chainlet.
        environment: The environment that the chainlet is deployed in.
          None if the chainlet is not associated with an environment.
    """

    data_dir: Optional[pathlib.Path] = None
    chainlet_to_service: Mapping[str, DeployedServiceDescriptor]
    secrets: MappingNoIter[str, str]
    environment: Optional[Environment] = None

    def get_service_descriptor(self, chainlet_name: str) -> DeployedServiceDescriptor:
        if chainlet_name not in self.chainlet_to_service:
            raise MissingDependencyError(f"{chainlet_name}")
        return self.chainlet_to_service[chainlet_name]

    def get_baseten_api_key(self) -> str:
        if self.secrets is None:
            raise ChainsRuntimeError(
                f"Secrets not set in `{self.__class__.__name__}` object."
            )
        error_msg = (
            "For using chains, it is required to setup a an API key with name "
            f"`{BASETEN_API_SECRET_NAME}` on Baseten to allow chain Chainlet to "
            "call other Chainlets. For local execution, secrets can be provided "
            "to `run_local`."
        )
        if BASETEN_API_SECRET_NAME not in self.secrets:
            raise MissingDependencyError(error_msg)

        api_key = self.secrets[BASETEN_API_SECRET_NAME]
        if api_key == SECRET_DUMMY:
            raise MissingDependencyError(
                f"{error_msg}. Retrieved dummy value of `{api_key}`."
            )
        return api_key


class TrussMetadata(SafeModel):
    """Plugin for the truss config (in config["model_metadata"]["chains_metadata"])."""

    chainlet_to_service: Mapping[str, ServiceDescriptor]


class ABCChainlet(abc.ABC):
    remote_config: ClassVar[RemoteConfig] = RemoteConfig()
    # `meta_data` is not shared between subclasses, each has an isolated copy.
    meta_data: ClassVar[ChainletMetadata] = ChainletMetadata()
    _framework_config: ClassVar[FrameworkConfig]

    @classmethod
    def has_custom_init(cls) -> bool:
        return cls.__init__ is not object.__init__

    @classproperty
    @classmethod
    def name(cls) -> str:
        return cls.__name__

    @classproperty
    @classmethod
    def display_name(cls) -> str:
        return cls.remote_config.name or cls.name

    @classproperty
    @classmethod
    def supports_dependencies(cls) -> bool:
        return cls._framework_config.supports_dependencies

    @classproperty
    @classmethod
    def entity_type(cls) -> EntityType:
        return cls._framework_config.entity_type

    @classproperty
    @classmethod
    def endpoint_method_name(cls) -> str:
        return cls._framework_config.endpoint_method_name

    # Cannot add this abstract method to API, because we want to allow arbitrary
    # arg/kwarg names and specifying any function signature here would give type errors
    # @abc.abstractmethod
    # def run_remote(self, *args, **kwargs) -> Any:
    #     ...


class TypeDescriptor(SafeModelNonSerializable):
    """For describing I/O types of Chainlets."""

    raw: Any  # The raw type annotation object (could be a type or GenericAlias).

    @property
    def is_pydantic(self) -> bool:
        return (
            isinstance(self.raw, type)
            and not isinstance(self.raw, GenericAlias)
            and issubclass(self.raw, pydantic.BaseModel)
        )

    @property
    def has_pydantic_args(self):
        origin = get_origin(self.raw)
        if not origin:
            return False
        args = get_args(self.raw)
        return any(
            isinstance(arg, type) and issubclass(arg, pydantic.BaseModel)
            for arg in args
        )

    @property
    def is_websocket(self) -> bool:
        return self.raw == WebSocketProtocol


class StreamingTypeDescriptor(TypeDescriptor):
    origin_type: type
    arg_type: type

    @property
    def is_string(self) -> bool:
        return self.arg_type is str

    @property
    def is_pydantic(self) -> bool:
        return False


class InputArg(SafeModelNonSerializable):
    name: str
    type: TypeDescriptor
    is_optional: bool


class EndpointAPIDescriptor(SafeModelNonSerializable):
    name: str = RUN_REMOTE_METHOD_NAME
    input_args: list[InputArg]
    output_types: list[TypeDescriptor]
    is_async: bool
    is_streaming: bool

    @property
    def streaming_type(self) -> StreamingTypeDescriptor:
        if (
            not self.is_streaming
            or len(self.output_types) != 1
            or not isinstance(self.output_types[0], StreamingTypeDescriptor)
        ):
            raise ValueError(f"{self} is not a streaming endpoint.")
        return cast(StreamingTypeDescriptor, self.output_types[0])

    @property
    def is_websocket(self):
        return any(arg.type.is_websocket for arg in self.input_args)

    @property
    def has_pydantic_input(self) -> bool:
        return not self.is_websocket

    @property
    def has_pydantic_output(self) -> bool:
        return not (self.is_streaming or self.is_websocket)

    @property
    def is_baseten_llm(self) -> bool:
        return any(arg.type.raw == baseten_llm.ModelInput for arg in self.input_args)


class DependencyDescriptor(SafeModelNonSerializable):
    chainlet_cls: Type[ABCChainlet]
    options: RPCOptions

    @property
    def name(self) -> str:
        return self.chainlet_cls.name

    @property
    def display_name(self) -> str:
        return self.chainlet_cls.display_name


class HealthCheckAPIDescriptor(SafeModelNonSerializable):
    name: str = HEALTH_CHECK_METHOD_NAME
    is_async: bool


class ChainletAPIDescriptor(SafeModelNonSerializable):
    chainlet_cls: Type[ABCChainlet]
    src_path: str
    has_context: bool
    dependencies: Mapping[str, DependencyDescriptor]
    endpoint: EndpointAPIDescriptor
    health_check: Optional[HealthCheckAPIDescriptor]

    def __hash__(self) -> int:
        return hash(self.chainlet_cls)

    @property
    def name(self) -> str:
        return self.chainlet_cls.name

    @property
    def display_name(self) -> str:
        return self.chainlet_cls.display_name


class StackFrame(SafeModel):
    filename: str
    lineno: Optional[int]
    name: str
    line: Optional[str]

    @classmethod
    def from_frame_summary(cls, frame: traceback.FrameSummary):
        return cls(
            filename=frame.filename,
            lineno=frame.lineno,
            name=frame.name,
            line=frame.line,
        )

    def to_frame_summary(self) -> traceback.FrameSummary:
        return traceback.FrameSummary(
            filename=self.filename, lineno=self.lineno, name=self.name, line=self.line
        )


class RemoteErrorDetail(SafeModel):
    """When a remote chainlet raises an exception, this pydantic model contains
    information about the error and stack trace and is included in JSON form in the
    error response.
    """

    exception_cls_name: str
    exception_module_name: Optional[str]
    exception_message: str
    user_stack_trace: list[StackFrame]

    def _to_stack_summary(self) -> traceback.StackSummary:
        return traceback.StackSummary.from_list(
            frame.to_frame_summary() for frame in self.user_stack_trace
        )

    def format(self) -> str:
        """Format the error for printing, similar to how Python formats exceptions
        with stack traces."""
        stack = "".join(traceback.format_list(self._to_stack_summary()))
        exc_info = (
            f"\n(Exception class defined in `{self.exception_module_name}`.)"
            if self.exception_module_name
            else ""
        )
        error = (
            f"Chainlet-Traceback (most recent call last):\n"
            f"{stack}{self.exception_cls_name}: {self.exception_message}{exc_info}"
        )
        return error


class GenericRemoteException(Exception): ...


########################################################################################


class PushOptions(SafeModelNonSerializable):
    chain_name: str
    only_generate_trusses: bool = False


class PushOptionsBaseten(PushOptions):
    remote: str
    publish: bool
    environment: Optional[str]

    @classmethod
    def create(
        cls,
        chain_name: str,
        publish: bool,
        promote: Optional[bool],
        only_generate_trusses: bool,
        remote: str,
        environment: Optional[str] = None,
    ) -> "PushOptionsBaseten":
        if promote and not environment:
            environment = PRODUCTION_ENVIRONMENT_NAME
        if environment:
            publish = True
        return PushOptionsBaseten(
            remote=remote,
            chain_name=chain_name,
            publish=publish,
            only_generate_trusses=only_generate_trusses,
            environment=environment,
        )


class PushOptionsLocalDocker(PushOptions):
    # Local docker-to-docker requests don't need auth, but we need to set a
    # value different from `SECRET_DUMMY` to not trigger the check that the secret
    # is unset. Additionally, if local docker containers make calls to models deployed
    # on baseten, a real API key must be provided (i.e. the default must be overridden).
    baseten_chain_api_key: str = "docker_dummy_key"
    # If enabled, chains code is copied from the local package into `/app/truss_chains`
    # in the docker image (which takes precedence over potential pip/site-packages).
    # This should be used for integration tests or quick local dev loops.
    use_local_chains_src: bool = False


class WebSocketProtocol(Protocol):
    """Describes subset of starlette/fastAPIs websocket interface that we expose."""

    headers: Mapping[str, str]

    async def close(self, code: int = 1000, reason: Optional[str] = None) -> None: ...

    async def receive_text(self) -> str: ...
    async def receive_bytes(self) -> bytes: ...
    async def receive_json(self) -> Any: ...

    async def send_text(self, data: str) -> None: ...
    async def send_bytes(self, data: bytes) -> None: ...
    async def send_json(self, data: Any) -> None: ...

    def iter_text(self) -> AsyncIterator[str]: ...
    def iter_bytes(self) -> AsyncIterator[bytes]: ...
    def iter_json(self) -> AsyncIterator[Any]: ...



import logging
import os
import sys
from dataclasses import _MISSING_TYPE, dataclass, field, fields
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, TypeVar

import yaml

from truss.base.constants import HTTP_PUBLIC_BLOB_BACKEND
from truss.base.custom_types import ModelFrameworkType
from truss.base.errors import ValidationError
from truss.base.trt_llm_config import (
    TRTLLMConfiguration,
    TrussTRTLLMBuildConfiguration,
    TrussTRTLLMQuantizationType,
)
from truss.base.validation import (
    validate_cpu_spec,
    validate_memory_spec,
    validate_node_count,
    validate_python_executable_path,
    validate_secret_name,
    validate_secret_to_path_mapping,
)
from truss.util.requirements import parse_requirement_string

DEFAULT_MODEL_FRAMEWORK_TYPE = ModelFrameworkType.CUSTOM
DEFAULT_MODEL_TYPE = "Model"
DEFAULT_MODEL_MODULE_DIR = "model"
DEFAULT_BUNDLED_PACKAGES_DIR = "packages"
DEFAULT_MODEL_CLASS_FILENAME = "model.py"
DEFAULT_MODEL_CLASS_NAME = "Model"
DEFAULT_TRUSS_STRUCTURE_VERSION = "2.0"
DEFAULT_MODEL_INPUT_TYPE = "Any"
DEFAULT_PYTHON_VERSION = "py39"
DEFAULT_DATA_DIRECTORY = "data"
DEFAULT_EXAMPLES_FILENAME = "examples.yaml"
DEFAULT_SPEC_VERSION = "2.0"
DEFAULT_PREDICT_CONCURRENCY = 1
DEFAULT_STREAMING_RESPONSE_READ_TIMEOUT = 60
DEFAULT_ENABLE_TRACING_DATA = False  # This should be in sync with tracing.py.
DEFAULT_ENABLE_DEBUG_LOGS = False
DEFAULT_IS_WEBSOCKET_ENDPOINT = False
DEFAULT_CPU = "1"
DEFAULT_MEMORY = "2Gi"
DEFAULT_USE_GPU = False

DEFAULT_BLOB_BACKEND = HTTP_PUBLIC_BLOB_BACKEND

VALID_PYTHON_VERSIONS = ["py38", "py39", "py310", "py311"]

X = TypeVar("X")
Y = TypeVar("Y")


def transform_optional(x: Optional[X], fn: Callable[[X], Optional[Y]]) -> Optional[Y]:
    if x is None:
        return None

    return fn(x)


logger = logging.getLogger(__name__)


class Accelerator(Enum):
    T4 = "T4"
    L4 = "L4"
    A10G = "A10G"
    V100 = "V100"
    A100 = "A100"
    H100 = "H100"
    H200 = "H200"
    H100_40GB = "H100_40GB"


@dataclass
class AcceleratorSpec:
    accelerator: Optional[Accelerator] = None
    count: int = 0

    def to_str(self) -> Optional[str]:
        if self.accelerator is None or self.count == 0:
            return None
        if self.count > 1:
            return f"{self.accelerator.value}:{self.count}"
        return self.accelerator.value

    @staticmethod
    def from_str(acc_spec: Optional[str]):
        if acc_spec is None:
            return AcceleratorSpec()
        parts = acc_spec.split(":")
        count = 1
        if len(parts) not in [1, 2]:
            raise ValidationError("`accelerator` does not match parsing requirements.")
        if len(parts) == 2:
            count = int(parts[1])
        try:
            acc = Accelerator[parts[0]]
        except KeyError as exc:
            raise ValidationError(f"Accelerator {acc_spec} not supported") from exc
        return AcceleratorSpec(accelerator=acc, count=count)


@dataclass
class ModelRepo:
    repo_id: str = ""
    revision: Optional[str] = None
    allow_patterns: Optional[List[str]] = None
    ignore_patterns: Optional[List[str]] = None

    @staticmethod
    def from_dict(d):
        repo_id = d.get("repo_id")
        if repo_id is None or repo_id == "":
            raise ValueError("Repo ID for Hugging Face model cannot be empty.")
        revision = d.get("revision", None)

        allow_patterns = d.get("allow_patterns", None)
        ignore_pattenrs = d.get("ignore_patterns", None)

        return ModelRepo(
            repo_id=repo_id,
            revision=revision,
            allow_patterns=allow_patterns,
            ignore_patterns=ignore_pattenrs,
        )

    def to_dict(self, verbose=False):
        data = {
            "repo_id": self.repo_id,
            "revision": self.revision,
            "allow_patterns": self.allow_patterns,
            "ignore_patterns": self.ignore_patterns,
        }

        if not verbose:
            # only show changed values
            data = {k: v for k, v in data.items() if v is not None}

        return data


@dataclass
class ModelCache:
    models: List[ModelRepo] = field(default_factory=list)

    @staticmethod
    def from_list(items: List[Dict[str, str]]) -> "ModelCache":
        return ModelCache([ModelRepo.from_dict(item) for item in items])

    def to_list(self, verbose=False) -> List[Dict[str, str]]:
        return [model.to_dict(verbose=verbose) for model in self.models]


@dataclass
class HealthChecks:
    restart_check_delay_seconds: Optional[int] = None
    restart_threshold_seconds: Optional[int] = None
    stop_traffic_threshold_seconds: Optional[int] = None

    @staticmethod
    def from_dict(d):
        return HealthChecks(
            restart_check_delay_seconds=d.get("restart_check_delay_seconds"),
            restart_threshold_seconds=d.get("restart_threshold_seconds"),
            stop_traffic_threshold_seconds=d.get("stop_traffic_threshold_seconds"),
        )

    def to_dict(self):
        return {
            "restart_check_delay_seconds": self.restart_check_delay_seconds,
            "restart_threshold_seconds": self.restart_threshold_seconds,
            "stop_traffic_threshold_seconds": self.stop_traffic_threshold_seconds,
        }


@dataclass
class Runtime:
    predict_concurrency: int = DEFAULT_PREDICT_CONCURRENCY
    streaming_read_timeout: int = DEFAULT_STREAMING_RESPONSE_READ_TIMEOUT
    enable_tracing_data: bool = DEFAULT_ENABLE_TRACING_DATA
    enable_debug_logs: bool = DEFAULT_ENABLE_DEBUG_LOGS
    health_checks: HealthChecks = field(default_factory=HealthChecks)
    is_websocket_endpoint: bool = DEFAULT_IS_WEBSOCKET_ENDPOINT

    @staticmethod
    def from_dict(d):
        predict_concurrency = d.get("predict_concurrency", DEFAULT_PREDICT_CONCURRENCY)
        num_workers = d.get("num_workers", 1)
        if num_workers != 1:
            raise ValueError(
                "After truss 0.9.49 only 1 worker per server is allowed. "
                "For concurrency utilize asyncio, autoscaling replicas "
                "and as a last resort thread/process pools inside the "
                "truss model."
            )
        streaming_read_timeout = d.get(
            "streaming_read_timeout", DEFAULT_STREAMING_RESPONSE_READ_TIMEOUT
        )
        enable_tracing_data = d.get("enable_tracing_data", DEFAULT_ENABLE_TRACING_DATA)
        enable_debug_logs = d.get("enable_debug_logs", DEFAULT_ENABLE_DEBUG_LOGS)
        health_checks = HealthChecks.from_dict(d.get("health_checks", {}))
        is_websocket_endpoint = d.get(
            "is_websocket_endpoint", DEFAULT_IS_WEBSOCKET_ENDPOINT
        )

        return Runtime(
            predict_concurrency=predict_concurrency,
            streaming_read_timeout=streaming_read_timeout,
            enable_tracing_data=enable_tracing_data,
            enable_debug_logs=enable_debug_logs,
            health_checks=health_checks,
            is_websocket_endpoint=is_websocket_endpoint,
        )

    def to_dict(self):
        return {
            "predict_concurrency": self.predict_concurrency,
            "streaming_read_timeout": self.streaming_read_timeout,
            "enable_tracing_data": self.enable_tracing_data,
            "health_checks": self.health_checks.to_dict(),
        }


class ModelServer(Enum):
    """
    To determine the image builder path for trusses built from alternative server backends.
    This enum is also used to gate development deployments to BasetenRemote
    https://github.com/basetenlabs/truss/blob/7505c17a2ddd4a6fa626b9126772999dc8f3fa86/truss/remote/baseten/remote.py#L56-L57
    """

    TrussServer = "TrussServer"
    TRT_LLM = "TRT_LLM"


@dataclass
class Build:
    model_server: ModelServer = ModelServer.TrussServer
    arguments: Dict = field(default_factory=dict)
    secret_to_path_mapping: Dict = field(default_factory=dict)

    @staticmethod
    def from_dict(d):
        model_server = ModelServer[d.get("model_server", "TrussServer")]
        arguments = d.get("arguments", {})
        secret_to_path_mapping = d.get("secret_to_path_mapping", {})
        validate_secret_to_path_mapping(secret_to_path_mapping)
        if not isinstance(secret_to_path_mapping, dict):
            raise ValueError(
                "Please pass a valid mapping for `secret_to_path_mapping`."
            )

        return Build(
            model_server=model_server,
            arguments=arguments,
            secret_to_path_mapping=secret_to_path_mapping,
        )

    def to_dict(self):
        return obj_to_dict(self)


@dataclass
class Resources:
    cpu: str = DEFAULT_CPU
    memory: str = DEFAULT_MEMORY
    use_gpu: bool = DEFAULT_USE_GPU
    accelerator: AcceleratorSpec = field(default_factory=AcceleratorSpec)
    node_count: Optional[int] = None

    @staticmethod
    def from_dict(d):
        cpu = d.get("cpu", DEFAULT_CPU)
        validate_cpu_spec(cpu)
        memory = d.get("memory", DEFAULT_MEMORY)
        validate_memory_spec(memory)
        accelerator = AcceleratorSpec.from_str((d.get("accelerator", None)))
        use_gpu = d.get("use_gpu", DEFAULT_USE_GPU)
        if accelerator.accelerator is not None:
            use_gpu = True

        r = Resources(cpu=cpu, memory=memory, use_gpu=use_gpu, accelerator=accelerator)

        # only add node_count if not None. This helps keep
        # config generated by truss init concise.
        node_count = d.get("node_count")
        validate_node_count(node_count)
        r.node_count = node_count

        return r

    def to_dict(self):
        d = {
            "cpu": self.cpu,
            "memory": self.memory,
            "use_gpu": self.use_gpu,
            "accelerator": self.accelerator.to_str(),
        }
        if self.node_count is not None:
            d["node_count"] = self.node_count
        return d


@dataclass
class ExternalDataItem:
    """A piece of remote data, to be made available to the Truss at serving time.

    Remote data is downloaded and stored under Truss's data directory. Care should be taken
    to avoid conflicts. This will get precedence if there's overlap.
    """

    # Url to download the data from.
    # Currently only files are allowed.
    url: str

    # This should be path relative to data directory. This is where the remote
    # file will be downloaded.
    local_data_path: str

    # The backend used to download files
    backend: str = DEFAULT_BLOB_BACKEND

    # A name can be given to a data item for readability purposes. It's not used
    # in the download process.
    name: Optional[str] = None

    @staticmethod
    def from_dict(d: Dict[str, str]) -> "ExternalDataItem":
        url = d.get("url")
        if url is None or url == "":
            raise ValueError("URL of an external data item cannot be empty")
        local_data_path = d.get("local_data_path")
        if local_data_path is None or local_data_path == "":
            raise ValueError(
                "The `local_data_path` field of an external data item cannot be empty"
            )

        item = ExternalDataItem(
            url=d["url"],
            local_data_path=d["local_data_path"],
            name=d.get("name"),
            backend=d.get("backend", DEFAULT_BLOB_BACKEND),
        )
        return item

    def to_dict(self):
        d = {
            "url": self.url,
            "local_data_path": self.local_data_path,
            "backend": self.backend,
        }
        if self.name is not None:
            d["name"] = self.name
        return d


@dataclass
class ExternalData:
    """[Experimental] External data is data that is not contained in the Truss folder.

    Typically this will be data stored remotely. This data is guaranteed to be made
    available under the data directory of the truss.
    """

    items: List[ExternalDataItem]

    @staticmethod
    def from_list(items: List[Dict[str, str]]) -> "ExternalData":
        return ExternalData([ExternalDataItem.from_dict(item) for item in items])

    def to_list(self) -> List[Dict[str, str]]:
        return [item.to_dict() for item in self.items]


class DockerAuthType(Enum):
    """
    This enum will express all of the types of registry
    authentication we support.
    """

    GCP_SERVICE_ACCOUNT_JSON = "GCP_SERVICE_ACCOUNT_JSON"


@dataclass
class DockerAuthSettings:
    """
    Provides information about how to authenticate to the docker registry containing
    the custom base image.
    """

    auth_method: DockerAuthType
    secret_name: str
    registry: Optional[str] = ""

    @staticmethod
    def from_dict(d: Dict[str, str]):
        auth_method = d.get("auth_method")
        secret_name = d.get("secret_name")

        if auth_method:
            # Capitalize the auth method so that we support this field passed
            # as "gcs_service_account".
            auth_method = auth_method.upper()

        if (
            not secret_name
            or not auth_method
            or auth_method not in [auth_type.value for auth_type in DockerAuthType]
        ):
            raise ValueError("Please provide a `secret_name`, and valid `auth_method`")

        return DockerAuthSettings(
            auth_method=DockerAuthType[auth_method],
            secret_name=secret_name,
            registry=d.get("registry"),
        )

    def to_dict(self):
        return {
            "auth_method": self.auth_method.value,
            "secret_name": self.secret_name,
            "registry": self.registry,
        }


@dataclass
class BaseImage:
    image: str = ""
    python_executable_path: str = ""
    docker_auth: Optional[DockerAuthSettings] = None

    @staticmethod
    def from_dict(d):
        image = d.get("image", "")
        python_executable_path = d.get("python_executable_path", "")
        docker_auth = d.get("docker_auth")
        validate_python_executable_path(python_executable_path)
        return BaseImage(
            image=image,
            python_executable_path=python_executable_path,
            docker_auth=(
                DockerAuthSettings.from_dict(docker_auth) if docker_auth else None
            ),
        )

    def to_dict(self):
        return {
            "image": self.image,
            "python_executable_path": self.python_executable_path,
            "docker_auth": transform_optional(
                self.docker_auth, lambda docker_auth: docker_auth.to_dict()
            ),
        }


@dataclass
class DockerServer:
    start_command: str
    server_port: int
    predict_endpoint: str
    readiness_endpoint: str
    liveness_endpoint: str

    @staticmethod
    def from_dict(d) -> "DockerServer":
        return DockerServer(
            start_command=d.get("start_command"),
            server_port=d.get("server_port"),
            predict_endpoint=d.get("predict_endpoint"),
            readiness_endpoint=d.get("readiness_endpoint"),
            liveness_endpoint=d.get("liveness_endpoint"),
        )

    def to_dict(self):
        return {
            "start_command": self.start_command,
            "server_port": self.server_port,
            "readiness_endpoint": self.readiness_endpoint,
            "liveness_endpoint": self.liveness_endpoint,
            "predict_endpoint": self.predict_endpoint,
        }


@dataclass
class TrussConfig:
    """
    `config.yaml` controls Truss config
    Args:
        description (str):
            Describe your model for documentation purposes.
        environment_variables (Dict[str, str]):
            <Warning>
            Do not store secret values directly in environment variables (or anywhere in the
            config file). See the `secrets` arg for information on properly managing secrets.
            </Warning>
            Any environment variables can be provided here as key value pairs and are exposed
            to the environment that the model executes in. Many Python libraries can be
            customized using environment variables, so this field can be quite handy in those
            scenarios.
            ```yaml
            environment_variables:
              ENVIRONMENT: Staging
              DB_URL: https://my_database.example.com/
            ```
        model_metadata (Dict[str, str]):
            Set any additional metadata in this catch-all field. The entire contents of the
            config file are available to the model at runtime, so this is a good place to
            store any custom information that model needs. For example, scikit-learn models
            include a flag here that indicates whether the model supports returning
            probabilities alongside predictions.
            ```yaml
            model_metadata:
              supports_predict_proba: true
            ```
        model_name (str):
            The model's name, for documentation purposes.
        requirements_file (str):
            Path of the requirements file with the required Python dependencies.
        requirements (List[str]):
            List the Python dependencies that the model depends on. The requirements should
            be provided in the
            [pip requirements file format](https://pip.pypa.io/en/stable/reference/requirements-file-format/),
            but as a yaml list. These requirements are installed after the ones from `requirements_file`.
            We strongly recommend pinning versions in your requirements.
            ```yaml
            requirements:
            - scikit-learn==1.0.2
            - threadpoolctl==3.0.0
            - joblib==1.1.0
            - numpy==1.20.3
            - scipy==1.7.3
            ```
        resources (Dict[str, str]):
            Specify model server runtime resources such as CPU, RAM and GPU.
            ```yaml
            resources:
              cpu: "3"
              memory: 14Gi
              use_gpu: true
              accelerator: A10G
              node_count: 2
            ```
        secrets (Dict[str, str]):
            <Warning>
            This field can be used to specify the keys for such secrets and dummy default
            values. ***Never store actual secret values in the config***. Dummy default
            values are instructive of what the actual values look like and thus act as
            documentation of the format.
            </Warning>
            A model may depend on certain secret values that can't be bundled with the model
            and need to be bound securely at runtime. For example, a model may need to download
            information from s3 and may need access to AWS credentials for that.
            ```yaml
            secrets:
              hf_access_token: "ACCESS TOKEN"
            ```
        system_packages (List[str]):
            Specify any system packages that you would typically install using `apt` on a Debian operating system.
            ```yaml
            system_packages:
            - ffmpeg
            - libsm6
            - libxext6
            ```
    Returns:
        `config.yaml` file which can be updated
    """

    model_framework: ModelFrameworkType = DEFAULT_MODEL_FRAMEWORK_TYPE
    model_type: str = DEFAULT_MODEL_TYPE
    model_name: Optional[str] = None

    model_module_dir: str = DEFAULT_MODEL_MODULE_DIR
    model_class_filename: str = DEFAULT_MODEL_CLASS_FILENAME
    model_class_name: str = DEFAULT_MODEL_CLASS_NAME

    data_dir: str = DEFAULT_DATA_DIRECTORY
    external_data: Optional[ExternalData] = None

    # Python types for what the model expects as input
    input_type: str = DEFAULT_MODEL_INPUT_TYPE
    model_metadata: Dict[str, Any] = field(default_factory=dict)
    requirements_file: Optional[str] = None
    requirements: List[str] = field(default_factory=list)
    system_packages: List[str] = field(default_factory=list)
    environment_variables: Dict[str, str] = field(default_factory=dict)
    resources: Resources = field(default_factory=Resources)
    runtime: Runtime = field(default_factory=Runtime)
    build: Build = field(default_factory=Build)
    python_version: str = DEFAULT_PYTHON_VERSION
    examples_filename: str = DEFAULT_EXAMPLES_FILENAME
    secrets: Dict[str, str] = field(default_factory=dict)
    description: Optional[str] = None
    bundled_packages_dir: str = DEFAULT_BUNDLED_PACKAGES_DIR
    external_package_dirs: List[str] = field(default_factory=list)
    live_reload: bool = False
    apply_library_patches: bool = True
    # spec_version is a version string
    spec_version: str = DEFAULT_SPEC_VERSION
    base_image: Optional[BaseImage] = None
    docker_server: Optional[DockerServer] = None
    model_cache: ModelCache = field(default_factory=ModelCache)
    trt_llm: Optional[TRTLLMConfiguration] = None
    build_commands: List[str] = field(default_factory=list)
    use_local_chains_src: bool = False

    @property
    def canonical_python_version(self) -> str:
        return {"py311": "3.11", "py310": "3.10", "py39": "3.9", "py38": "3.8"}[
            self.python_version
        ]

    @property
    def parsed_trt_llm_build_configs(self) -> List[TrussTRTLLMBuildConfiguration]:
        if self.trt_llm:
            if self.trt_llm.build.speculator and self.trt_llm.build.speculator.build:
                return [self.trt_llm.build, self.trt_llm.build.speculator.build]
            return [self.trt_llm.build]
        return []

    @staticmethod
    def from_dict(d):
        config = TrussConfig(
            spec_version=d.get("spec_version", DEFAULT_SPEC_VERSION),
            model_type=d.get("model_type", DEFAULT_MODEL_TYPE),
            model_framework=ModelFrameworkType(
                d.get("model_framework", DEFAULT_MODEL_FRAMEWORK_TYPE.value)
            ),
            model_module_dir=d.get("model_module_dir", DEFAULT_MODEL_MODULE_DIR),
            model_class_filename=d.get(
                "model_class_filename", DEFAULT_MODEL_CLASS_FILENAME
            ),
            model_class_name=d.get("model_class_name", DEFAULT_MODEL_CLASS_NAME),
            data_dir=d.get("data_dir", DEFAULT_DATA_DIRECTORY),
            input_type=d.get("input_type", DEFAULT_MODEL_INPUT_TYPE),
            model_metadata=d.get("model_metadata", {}),
            requirements_file=d.get("requirements_file", None),
            requirements=d.get("requirements", []),
            system_packages=d.get("system_packages", []),
            environment_variables=_handle_env_vars(d.get("environment_variables", {})),
            resources=Resources.from_dict(d.get("resources", {})),
            runtime=Runtime.from_dict(d.get("runtime", {})),
            build=Build.from_dict(d.get("build", {})),
            python_version=d.get("python_version", DEFAULT_PYTHON_VERSION),
            model_name=d.get("model_name", None),
            examples_filename=d.get("examples_filename", DEFAULT_EXAMPLES_FILENAME),
            secrets=d.get("secrets", {}),
            description=d.get("description", None),
            bundled_packages_dir=d.get(
                "bundled_packages_dir", DEFAULT_BUNDLED_PACKAGES_DIR
            ),
            external_package_dirs=d.get("external_package_dirs", []),
            live_reload=d.get("live_reload", False),
            apply_library_patches=d.get("apply_library_patches", True),
            external_data=transform_optional(
                d.get("external_data"), ExternalData.from_list
            ),
            base_image=transform_optional(d.get("base_image"), BaseImage.from_dict),
            docker_server=transform_optional(
                d.get("docker_server"), DockerServer.from_dict
            ),
            model_cache=transform_optional(
                d.get("model_cache") or d.get("hf_cache") or [],  # type: ignore
                ModelCache.from_list,
            ),
            trt_llm=transform_optional(
                d.get("trt_llm"), lambda x: (TRTLLMConfiguration(**x))
            ),
            build_commands=d.get("build_commands", []),
            use_local_chains_src=d.get("use_local_chains_src", False),
        )
        config.validate()
        return config

    def load_requirements_from_file(self, truss_dir: Path) -> List[str]:
        if self.requirements_file:
            requirements_path = truss_dir / self.requirements_file
            try:
                requirements = []
                with open(requirements_path) as f:
                    for line in f.readlines():
                        parsed_line = parse_requirement_string(line)
                        if parsed_line:
                            requirements.append(parsed_line)
                return requirements
            except Exception as e:
                logger.exception(
                    f"failed to read requirements file: {self.requirements_file}"
                )
                raise e
        return []

    @staticmethod
    def load_requirements_file_from_filepath(yaml_path: Path) -> List[str]:
        config = TrussConfig.from_yaml(yaml_path)
        return config.load_requirements_from_file(yaml_path.parent)

    @staticmethod
    def from_yaml(yaml_path: Path):
        if not os.path.isfile(yaml_path):
            raise ValueError(f"Expected a truss configuration file at {yaml_path}")
        with yaml_path.open() as yaml_file:
            raw_data = yaml.safe_load(yaml_file) or {}
            if "hf_cache" in raw_data:
                logger.warning(
                    """Warning: `hf_cache` is deprecated in favor of `model_cache`.
                    Everything will run as before, but if you are pulling weights from S3 or GCS, they will be
                    stored at /app/model_cache instead of /app/hf_cache as before."""
                )
            return TrussConfig.from_dict(raw_data)

    def write_to_yaml_file(self, path: Path, verbose: bool = True):
        with path.open("w") as config_file:
            yaml.dump(self.to_dict(verbose=verbose), config_file)

    def to_dict(self, verbose: bool = True):
        return obj_to_dict(self, verbose=verbose)

    def clone(self):
        return TrussConfig.from_dict(self.to_dict())

    def _validate_trt_llm_config(self) -> None:
        if self.trt_llm:
            if (
                self.trt_llm.build.quantization_type
                is TrussTRTLLMQuantizationType.WEIGHTS_ONLY_INT8
                and self.resources.accelerator.accelerator is Accelerator.A100
            ):
                raise ValueError(
                    "Weight only int8 quantization on A100 accelerators is not currently supported"
                )
            elif self.trt_llm.build.quantization_type in [
                TrussTRTLLMQuantizationType.FP8,
                TrussTRTLLMQuantizationType.FP8_KV,
            ] and self.resources.accelerator.accelerator not in [
                Accelerator.H100,
                Accelerator.H100_40GB,
                Accelerator.L4,
            ]:
                raise ValueError(
                    "FP8 quantization is only supported on L4 and H100 accelerators"
                )
            tensor_parallel_count = self.trt_llm.build.tensor_parallel_count

            if tensor_parallel_count != self.resources.accelerator.count:
                raise ValueError(
                    "Tensor parallelism and GPU count must be the same for TRT-LLM"
                )

    def validate(self):
        if self.python_version not in VALID_PYTHON_VERSIONS:
            raise ValueError(
                f"Please ensure that `python_version` is one of {VALID_PYTHON_VERSIONS}"
            )

        if not isinstance(self.secrets, dict):
            raise ValueError(
                "Please ensure that `secrets` is a mapping of the form:\n"
                "```\n"
                "secrets:\n"
                '  secret1: "some default value"\n'
                '  secret2: "some other default value"\n'
                "```"
            )
        for secret_name in self.secrets:
            validate_secret_name(secret_name)

        if self.requirements and self.requirements_file:
            raise ValueError(
                "Please ensure that only one of `requirements` and `requirements_file` is specified"
            )
        self._validate_trt_llm_config()


def _handle_env_vars(env_vars: Dict[str, Any]) -> Dict[str, str]:
    new_env_vars = {}
    for k, v in env_vars.items():
        if isinstance(v, bool):
            new_env_vars[k] = str(v).lower()
        else:
            new_env_vars[k] = str(v)
    return new_env_vars


DATACLASS_TO_REQ_KEYS_MAP = {
    Resources: {"accelerator", "cpu", "memory", "use_gpu"},
    Runtime: {"predict_concurrency"},
    Build: {"model_server"},
    TrussConfig: {
        "environment_variables",
        "external_package_dirs",
        "model_metadata",
        "model_name",
        "python_version",
        "requirements",
        "resources",
        "secrets",
        "system_packages",
        "build_commands",
    },
    BaseImage: {"image", "python_executable_path"},
}


def obj_to_dict(obj, verbose: bool = False):
    """
    This function serializes a given object (usually starting with a TrussConfig) and
    only keeps required keys or ones changed by the user manually. This simplifies the config.yml.
    """
    required_keys = DATACLASS_TO_REQ_KEYS_MAP[type(obj)]
    d = {}
    for f in fields(obj):
        field_name = f.name
        field_default_value = f.default
        field_default_factory = f.default_factory

        field_curr_value = getattr(obj, f.name)

        expected_default_value = None
        if not isinstance(field_default_value, _MISSING_TYPE):
            expected_default_value = field_default_value
        else:
            expected_default_value = field_default_factory()  # type: ignore

        should_add_to_dict = (
            expected_default_value != field_curr_value or field_name in required_keys
        )

        if verbose or should_add_to_dict:
            if isinstance(field_curr_value, tuple(DATACLASS_TO_REQ_KEYS_MAP.keys())):
                d[field_name] = obj_to_dict(field_curr_value, verbose=verbose)
            elif isinstance(field_curr_value, AcceleratorSpec):
                d[field_name] = field_curr_value.to_str()
            elif isinstance(field_curr_value, Enum):
                d[field_name] = field_curr_value.value
            elif isinstance(field_curr_value, ExternalData):
                d["external_data"] = transform_optional(
                    field_curr_value, lambda data: data.to_list()
                )
            elif isinstance(field_curr_value, ModelCache):
                d["model_cache"] = transform_optional(
                    field_curr_value, lambda data: data.to_list(verbose=verbose)
                )
            elif isinstance(field_curr_value, TRTLLMConfiguration):
                d["trt_llm"] = transform_optional(
                    field_curr_value, lambda data: data.to_json_dict(verbose=verbose)
                )
            elif isinstance(field_curr_value, BaseImage):
                d["base_image"] = transform_optional(
                    field_curr_value, lambda data: data.to_dict()
                )
            elif isinstance(field_curr_value, DockerServer):
                d["docker_server"] = transform_optional(
                    field_curr_value, lambda data: data.to_dict()
                )
            elif isinstance(field_curr_value, DockerAuthSettings):
                d["docker_auth"] = transform_optional(
                    field_curr_value, lambda data: data.to_dict()
                )
            elif isinstance(field_curr_value, HealthChecks):
                d["health_checks"] = transform_optional(
                    field_curr_value, lambda data: data.to_dict()
                )
            else:
                d[field_name] = field_curr_value

    return d


def _infer_python_version() -> str:
    return f"py{sys.version_info.major}{sys.version_info.minor}"


def map_local_to_supported_python_version() -> str:
    return _map_to_supported_python_version(_infer_python_version())


def _map_to_supported_python_version(python_version: str) -> str:
    """Map python version to truss supported python version.

    Currently, it maps any versions greater than 3.11 to 3.11.

    Args:
        python_version: in the form py[major_version][minor_version] e.g. py39,
          py310.
    """
    python_major_version = int(python_version[2:3])
    python_minor_version = int(python_version[3:])

    if python_major_version != 3:
        raise NotImplementedError("Only python version 3 is supported")

    if python_minor_version > 11:
        logger.info(
            f"Mapping python version {python_major_version}.{python_minor_version}"
            " to 3.11, the highest version that Truss currently supports."
        )
        return "py311"

    if python_minor_version < 8:
        raise ValueError(
            f"Mapping python version {python_major_version}.{python_minor_version}"
            " to 3.8, the lowest version that Truss currently supports."
        )

    return python_version



import pathlib
from typing import Optional

from truss.templates.shared import secrets_resolver

from truss_chains import definitions
from truss_chains.remote_chainlet import utils


class TrussChainletModel:
    _context: definitions.DeploymentContext
    _chainlet: definitions.ABCChainlet

    def __init__(
        self,
        config: dict,
        data_dir: pathlib.Path,
        secrets: secrets_resolver.Secrets,
        # TODO: Remove the default value once all truss versions are synced up.
        environment: Optional[dict] = None,
    ) -> None:
        truss_metadata: definitions.TrussMetadata = (
            definitions.TrussMetadata.model_validate(
                config["model_metadata"][definitions.TRUSS_CONFIG_CHAINS_KEY]
            )
        )
        deployment_environment: Optional[definitions.Environment] = (
            definitions.Environment.model_validate(environment) if environment else None
        )
        chainlet_to_deployed_service = utils.populate_chainlet_service_predict_urls(
            truss_metadata.chainlet_to_service
        )

        self._context = definitions.DeploymentContext(
            chainlet_to_service=chainlet_to_deployed_service,
            secrets=secrets,
            data_dir=data_dir,
            environment=deployment_environment,
        )

    # Below illustrated code will be added by code generation.

    # def load(self) -> None:
    #     logging.info(f"Loading Chainlet `TextToNum`.")
    #     self._chainlet = itest_chain.TextToNum(
    #         replicator=stub.factory(TextReplicator, self._context),
    #         side_effect=stub.factory(SideEffectOnlySubclass, self._context),
    #     )
    #
    # If chainlet implements is_healthy:
    # def is_healthy(self) -> Optional[bool]:
    #     if hasattr(self, "_chainlet"):
    #         return self._chainlet.is_healthy()
    #
    # def predict(
    #     self, inputs: TextToNumInput, request: starlette.requests.Request
    # ) -> TextToNumOutput:
    #     with utils.predict_context(request.headers):
    #         result = self._chainlet.run_remote(**utils.pydantic_set_field_dict(inputs))
    #     return TextToNumOutput(result)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional

import requests

if TYPE_CHECKING:
    from rich import console as rich_console
from truss.truss_handle.truss_handle import TrussHandle


class RemoteUser:
    """Class to hold information about the remote user"""

    workspace_name: str
    user_email: str

    def __init__(self, workspace_name: str, user_email: str):
        self.workspace_name = workspace_name
        self.user_email = user_email


class TrussService(ABC):
    """
    Define the abstract base class for a TrussService.

    A TrussService interacts with a service at a given URL and can either be in
    draft or non-draft mode.

    Attributes:
        _service_url: The URL of the service.
        _is_draft: A boolean indicating if the service is in draft mode.

    Args:
        service_url: The URL of the service.
        is_draft: A boolean indicating if the service is in draft mode.
        **kwargs: Additional keyword arguments to initialize the service.

    """

    def __init__(self, service_url: str, is_draft: bool, **kwargs) -> None:
        self._service_url = service_url
        self._is_draft = is_draft

    def _send_request(
        self,
        url: str,
        method: str,
        headers: Optional[Dict] = None,
        data: Optional[Dict] = None,
        stream: Optional[bool] = False,
    ) -> Any:
        """
        Send a HTTP request.

        Args:
            url: The URL to send the request to.
            method: The HTTP method to use.
            headers: Optional dictionary of headers to include in the request.
            data: Optional dictionary of data to include in the request.

        Raises:
            ValueError: If an unsupported method is used or a POST request has no data.

        Returns:
            A Response object resulting from the request.

        """
        if not headers:
            headers = {}

        auth_header = self.authenticate()
        headers = {**headers, **auth_header}
        if method == "GET":
            response = requests.request(method, url, headers=headers, stream=stream)
        elif method == "POST":
            if data is None:
                raise ValueError("POST request must have data")
            response = requests.request(
                method, url, json=data, headers=headers, stream=stream
            )
        else:
            raise ValueError(f"Unsupported method: {method}")

        if response.status_code == 401:
            raise ValueError(
                f"Authentication failed with status code {response.status_code}"
            )

        return response

    @property
    def is_draft(self) -> bool:
        """
        Check if the service is in draft mode.

        Returns:
            A boolean indicating if the service is in draft mode.
        """
        return self._is_draft

    @abstractmethod
    def is_live(self) -> bool:
        """
        Check if the service is live.

        Sends a GET request to the root of the service and returns whether it
        is successful.

        Returns:
            A boolean indicating if the service is live.
        """
        liveness_url = f"{self._service_url}/"
        response = self._send_request(liveness_url, "GET", {})
        return response.status_code == 200

    @abstractmethod
    def is_ready(self) -> bool:
        """
        Check if the service is ready.

        Sends a GET request to the model path of the service and returns whether it
        is successful.

        Returns:
            A boolean indicating if the service is ready.
        """
        readiness_url = f"{self._service_url}/v1/models/model"
        response = self._send_request(readiness_url, "GET", {})
        return response.status_code == 200

    def predict(self, model_request_body: Dict) -> Any:
        """
        Send a prediction request to the service.

        Args:
            model_request_body: A dictionary representing the body of the
            prediction request.

        Returns:
            A Response object resulting from the prediction request.
        """
        return self._send_request(self.predict_url, "POST", data=model_request_body)

    def patch(self) -> None:
        """
        Patch the service. TrussServices in draft mode can be patched.
        """
        raise NotImplementedError

    @abstractmethod
    def authenticate(self) -> dict:
        """
        Authenticate to the service.

        This method should be implemented in subclasses and return a dictionary
        of headers to include in requests to the service with authentication
        information.
        """
        return {}

    @property
    @abstractmethod
    def logs_url(self) -> str:
        """
        Get the URL for the service logs.
        """
        pass

    @property
    @abstractmethod
    def predict_url(self) -> str:
        """
        Get the URL for the prediction endpoint.
        """
        pass

    @abstractmethod
    def poll_deployment_status(self, sleep_secs: int = 1) -> Iterator[str]:
        """
        Poll for a deployment status.
        """
        pass


class TrussRemote(ABC):
    """
    Define the abstract base class for a remote Truss service.

    A remote Truss service is a service that can push a TrussHandle to a remote
    location. The `push` and `authenticate` methods should be implemented in subclasses.

    Attributes:
        _remote_url: The URL of the remote service.

    Args:
        remote_url: The URL of the remote service.
        **kwargs: Additional keyword arguments needed to initialize the remote service.

    """

    def __init__(self, remote_url: str) -> None:
        self._remote_url = remote_url

    @property
    def remote_url(self) -> str:
        return self._remote_url

    @abstractmethod
    def push(self, truss_handle: TrussHandle, **kwargs) -> TrussService:
        """
        Push a TrussHandle to the remote service.

        This method should be implemented in subclasses and return a TrussService
        object for interacting with the remote service.

        Args:
            truss_handle: The TrussHandle to push to the remote service.
            **kwargs: Additional keyword arguments for the push operation.

        """

    @abstractmethod
    def whoami(self) -> RemoteUser:
        """
        Returns account information for the current user.

        This method should be implemented in subclasses and return a RemoteUser.


        """

    @abstractmethod
    def get_service(self, **kwargs) -> TrussService:
        """
        Get a TrussService object for interacting with the remote service.

        This method should be implemented in subclasses and return a TrussService
        object for interacting with the remote service. This support keyword args,
        so that implementing remotes can be flexible with how a service is instantiated.

        Args:
            **kwargs: Keyword arguments for the get_service operation.

        """

    @abstractmethod
    def sync_truss_to_dev_version_by_name(
        self,
        model_name: str,
        target_directory: str,
        console: "rich_console.Console",
        error_console: "rich_console.Console",
    ) -> None:
        """
        This method watches for changes to files in the `target_directory`,
        and syncs them to the development version of the model, identified
        by name.

        Args:
            model_name: The name of the model to sync to the dev version
            target_directory: The directory to sync the model to
            console: For printing informative output.
            error_console: For printing errors.
        """


@dataclass
class RemoteConfig:
    """Class to hold configs for various remotes"""

    name: str
    configs: Dict = field(default_factory=dict)



from unittest import mock

import pytest
import requests
from requests import Response
from truss.remote.baseten import custom_types as b10_types
from truss.remote.baseten.api import BasetenApi
from truss.remote.baseten.custom_types import ChainletDataAtomic, OracleData
from truss.remote.baseten.error import ApiError


@pytest.fixture
def mock_auth_service():
    auth_service = mock.Mock()
    auth_token = mock.Mock(headers=lambda: {"Authorization": "Api-Key token"})
    auth_service.authenticate.return_value = auth_token
    return auth_service


def mock_successful_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(return_value={"data": {"status": "success"}})
    return response


def mock_graphql_error_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(return_value={"errors": [{"message": "error"}]})
    return response


def mock_unsuccessful_response():
    response = Response()
    response.status_code = 400
    return response


def mock_create_model_version_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(
        return_value={
            "data": {
                "create_model_version_from_truss": {"model_version": {"id": "12345"}}
            }
        }
    )
    return response


def mock_create_model_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(
        return_value={
            "data": {"create_model_from_truss": {"model_version": {"id": "12345"}}}
        }
    )
    return response


def mock_create_development_model_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(
        return_value={
            "data": {"deploy_draft_truss": {"model_version": {"id": "12345"}}}
        }
    )
    return response


def mock_deploy_chain_deployment_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(
        return_value={
            "data": {
                "deploy_chain_atomic": {
                    "chain_deployment": {"id": "54321", "chain": {"id": "12345"}}
                }
            }
        }
    )
    return response


@pytest.fixture
def baseten_api(mock_auth_service):
    return BasetenApi("https://app.test.com", mock_auth_service)


@mock.patch("requests.post", return_value=mock_successful_response())
def test_post_graphql_query_success(mock_post, baseten_api):
    response_data = {"data": {"status": "success"}}

    result = baseten_api._post_graphql_query("sample_query_string")

    assert result == response_data


@mock.patch("requests.post", return_value=mock_graphql_error_response())
def test_post_graphql_query_error(mock_post, baseten_api):
    with pytest.raises(ApiError):
        baseten_api._post_graphql_query("sample_query_string")


@mock.patch("requests.post", return_value=mock_unsuccessful_response())
def test_post_requests_error(mock_post, baseten_api):
    with pytest.raises(requests.exceptions.HTTPError):
        baseten_api._post_graphql_query("sample_query_string")


@mock.patch("requests.post", return_value=mock_create_model_version_response())
def test_create_model_version_from_truss(mock_post, baseten_api):
    baseten_api.create_model_version_from_truss(
        "model_id",
        "s3key",
        "config_str",
        "semver_bump",
        False,
        "deployment_name",
        "production",
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]
    assert 'model_id: "model_id"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert {
        "trussUserEnv": b10_types.TrussUserEnv.collect().model_dump_json()
    } == mock_post.call_args[1]["json"]["variables"]
    assert "scale_down_old_production: true" in gql_mutation
    assert 'name: "deployment_name"' in gql_mutation
    assert 'environment_name: "production"' in gql_mutation


@mock.patch("requests.post", return_value=mock_create_model_version_response())
def test_create_model_version_from_truss_does_not_send_deployment_name_if_not_specified(
    mock_post, baseten_api
):
    baseten_api.create_model_version_from_truss(
        "model_id", "s3key", "config_str", "semver_bump", False, deployment_name=None
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]
    assert 'model_id: "model_id"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert {
        "trussUserEnv": b10_types.TrussUserEnv.collect().model_dump_json()
    } == mock_post.call_args[1]["json"]["variables"]
    assert "scale_down_old_production: true" in gql_mutation
    assert " name: " not in gql_mutation
    assert "environment_name: " not in gql_mutation


@mock.patch("requests.post", return_value=mock_create_model_version_response())
def test_create_model_version_from_truss_does_not_scale_old_prod_to_zero_if_keep_previous_prod_settings(
    mock_post, baseten_api
):
    baseten_api.create_model_version_from_truss(
        "model_id",
        "s3key",
        "config_str",
        "semver_bump",
        True,
        deployment_name=None,
        environment="staging",
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]
    assert 'model_id: "model_id"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert {
        "trussUserEnv": b10_types.TrussUserEnv.collect().model_dump_json()
    } == mock_post.call_args[1]["json"]["variables"]
    assert "scale_down_old_production: false" in gql_mutation
    assert " name: " not in gql_mutation
    assert 'environment_name: "staging"' in gql_mutation


@mock.patch("requests.post", return_value=mock_create_model_response())
def test_create_model_from_truss(mock_post, baseten_api):
    baseten_api.create_model_from_truss(
        "model_name",
        "s3key",
        "config_str",
        "semver_bump",
        deployment_name="deployment_name",
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]
    assert 'name: "model_name"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert {
        "trussUserEnv": b10_types.TrussUserEnv.collect().model_dump_json()
    } == mock_post.call_args[1]["json"]["variables"]
    assert 'version_name: "deployment_name"' in gql_mutation


@mock.patch("requests.post", return_value=mock_create_model_response())
def test_create_model_from_truss_does_not_send_deployment_name_if_not_specified(
    mock_post, baseten_api
):
    baseten_api.create_model_from_truss(
        "model_name", "s3key", "config_str", "semver_bump", deployment_name=None
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]
    assert 'name: "model_name"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert {
        "trussUserEnv": b10_types.TrussUserEnv.collect().model_dump_json()
    } == mock_post.call_args[1]["json"]["variables"]
    assert "version_name: " not in gql_mutation


@mock.patch("requests.post", return_value=mock_create_model_response())
def test_create_model_from_truss_with_allow_truss_download(mock_post, baseten_api):
    baseten_api.create_model_from_truss(
        "model_name", "s3key", "config_str", "semver_bump", allow_truss_download=False
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]
    assert 'name: "model_name"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert {
        "trussUserEnv": b10_types.TrussUserEnv.collect().model_dump_json()
    } == mock_post.call_args[1]["json"]["variables"]
    assert "allow_truss_download: false" in gql_mutation


@mock.patch("requests.post", return_value=mock_create_development_model_response())
def test_create_development_model_from_truss_with_allow_truss_download(
    mock_post, baseten_api
):
    baseten_api.create_development_model_from_truss(
        "model_name", "s3key", "config_str", allow_truss_download=False
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]
    assert 'name: "model_name"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert {
        "trussUserEnv": b10_types.TrussUserEnv.collect().model_dump_json()
    } == mock_post.call_args[1]["json"]["variables"]
    assert "allow_truss_download: false" in gql_mutation


@mock.patch("requests.post", return_value=mock_deploy_chain_deployment_response())
def test_deploy_chain_deployment(mock_post, baseten_api):
    baseten_api.deploy_chain_atomic(
        environment="production",
        chain_id="chain_id",
        dependencies=[],
        entrypoint=ChainletDataAtomic(
            name="chainlet-1",
            oracle=OracleData(
                model_name="model-1",
                s3_key="s3-key-1",
                encoded_config_str="encoded-config-str-1",
            ),
        ),
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]

    assert 'environment: "production"' in gql_mutation
    assert 'chain_id: "chain_id"' in gql_mutation
    assert "dependencies:" in gql_mutation
    assert "entrypoint:" in gql_mutation


@mock.patch("requests.post", return_value=mock_deploy_chain_deployment_response())
def test_deploy_chain_deployment_no_environment(mock_post, baseten_api):
    baseten_api.deploy_chain_atomic(
        chain_id="chain_id",
        dependencies=[],
        entrypoint=ChainletDataAtomic(
            name="chainlet-1",
            oracle=OracleData(
                model_name="model-1",
                s3_key="s3-key-1",
                encoded_config_str="encoded-config-str-1",
            ),
        ),
    )

    gql_mutation = mock_post.call_args[1]["json"]["query"]

    assert 'chain_id: "chain_id"' in gql_mutation
    assert "environment" not in gql_mutation
    assert "dependencies:" in gql_mutation
    assert "entrypoint:" in gql_mutation
